' Simple Sprite Draw
' written by Vegipete, November, 2023
' for PicoMiteVGA (and maybe others)
' Generates CSub sprite data as used by the BLIT MEMORY command
' Output is sent to the console, for copying into source code.
'
' Simple animations are possible by cycling through frames.
'
Option default integer
Option break 25 ' use ^Y instead of ^C to stop program (^C = Copy)
Option LCDPANEL NOCONSOLE
Dim cmap(15)=(0,1,6,7,8,9,14,15,2,3,4,5,10,11,12,13)
MODE 2 : CLS 0
Box 0,0,MM.HRes,MM.VRes,,RGB(myrtle)  ' show screen extents
FRAMEBUFFER layer 13
FRAMEBUFFER write L
CLS RGB(lilac)
FRAMEBUFFER write n

' Following is convenient for showing ascii code of keystrokes
'do:k=asc(inkey$):if k<>0 then:? k:endif:loop

Const MAX_X = 64 ' divisible by 16!
Const MAX_Y = 64
Dim size_x = 48
Dim size_y = 40
Dim sc = 5
Dim img(MAX_X-1,MAX_Y-1)
Dim undo1(MAX_X*MAX_Y/16)   ' undo buffer  16 pixels fit in 1 integer
' 2 or 3 or ? undo levels might be possible
Dim ccp(MAX_X/16-1,MAX_Y-1) ' 16 pixels per (64 bit) integer
ccpx = 0 : ccpy = 0   ' size of data in CCP buffer
Math set 0,img()
Math set 0,undo1()
Math set 0,ccp()
tmp = 0
undoready = 0   ' set to 1 after something that can be undone
zoomx = 70
zoomy = 20
sbx = 10  ' sprite selection box x
sby = 10  ' sprite selection box y
sbh = 10  ' sprite selection box horizontal size
sbv = 11  ' sprite selection box vertical size
rsx = 3 ' real size view x
rsy = 120 ' real size view y
hbx = 20 : hby = 8 : hbw = 280 : hbh = 215 ' help box
abx = 0 : aby = 0 : aex = size_x : aey = size_y ' animation begin and end
cn$ = "0123456789ABCDEF"
ltr$ = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
dgt$ = "0123456789"
Dim wss$(3) = (" 32 x 32 "," 48 x 40 "," 60 x 48 "," 64 x 64 ")
Dim wssx(3) = (-25,-25,25,25)
Dim wssj$(3) = ("RB","RT","LB","LT")
Dim ctab(15):For i = 0 To 15: Read ctab(i) : Next
Data RGB(BLACK),RGB(BLUE),RGB(GREEN),RGB(CYAN),RGB(RED)
Data RGB(MAGENTA),RGB(YELLOW),RGB(WHITE),RGB(MYRTLE)
Data RGB(COBALT) ,RGB(MIDGREEN),RGB(CERULEAN)
Data RGB(RUST), RGB(FUCHSIA),RGB(BROWN),RGB(LILAC)

'===================
' GetFileDialog Setup
' required setup for GetFileName dialog function
Const DIRCOUNT = 50   ' max number of sub-directories
Const FILCOUNT = 255  ' max number of files
Const FNAMELEN = 32   ' length of file names
Dim dir_dirs$(DIRCOUNT) length FNAMELEN  ' store list of directories
Dim dir_fils$(FILCOUNT) length FNAMELEN  ' store list of files
Dim dir_hist$(DIRCOUNT) length 8 ' store directory number visited along path
Dim d_shadow = RGB(blue)      ' dialog box colours
Dim d_frame =  RGB(cobalt)
Dim d_back  =  0
Dim d_fore  =  RGB(cyan)  ' hilite text colour
' end of setup for function
'===================

DrawTitle(1)
ShowZoomView
ShowPallet
cx = sbx : cy = sby
ShowCurCol(img(cx,cy))
ShowSSB
ShowCur
ShowValues
ShowRealSize

Do
  k = Asc(UCase$(Inkey$))
  If k Then
    Select Case k
      Case 130  ' left
        If cx > 0 Then HideCur : Inc cx,-1
      Case 131  ' right
        If cx < size_x-1 Then HideCur : Inc cx
      Case 128  ' up
        If cy > 0 Then HideCur : Inc cy,-1
      Case 129  ' down
        If cy < size_y-1 Then HideCur : Inc cy
      Case 48 To 57  ' digits 0 to 9
        img(cx,cy) = k-48
        DrawCell(cx,cy) 'ShowZoomView
      Case 65 To 70  ' letters A to F
        img(cx,cy) = k-55
        DrawCell(cx,cy) 'ShowZoomView
      Case 73   ' I up
        If sby > 0 Then HideSSB : Inc sby,-1
      Case 74   ' J left
        If sbx > 0 Then HideSSB : Inc sbx,-1
      Case 75   ' K down
        If sby+sbv < size_y Then HideSSB : Inc sby
      Case 76   ' L right
        If sbx+sbh < size_x Then HideSSB : Inc sbx
      Case 44   ' , narrower
        If sbh > 1 Then HideSSB : Inc sbh,-1
      Case 46   ' . wider
        If sbx+sbh < size_x Then HideSSB : Inc sbh
      Case 59   ' ; taller
        If sby+sbv < size_y Then HideSSB : Inc sbv
      Case 80   ' P shorter
        If sbv > 1 Then HideSSB : Inc sbv,-1
      Case 63,145 ' ?,F1 help
        ShowHelp
      Case 71   ' G generate CSub
        MakeCSub
      Case 72   ' H flip vertical
        FlipVertical
      Case 77   ' M flip horizontal
        FlipHorizontal
      Case 78   ' N flip diagonal
        FlipDiagonal
      Case 82   ' R resize worksheet
        tmp = GetSizeDialog(6-sc)
        If tmp = 0 Then size_x = 32 : size_y = 32 : sc = 6
        If tmp = 1 Then size_x = 48 : size_y = 40 : sc = 5
        If tmp = 2 Then size_x = 60 : size_y = 48 : sc = 4
        If tmp = 3 Then size_x = 64 : size_y = 64 : sc = 3
        FRAMEBUFFER write n
        Box zoomx,zoomy,249,210,,0,0   ' erase old zoom view
        Box rsx-2,rsy-2,67,67,,0,0     ' erase old resize view
        ShowZoomView
        If sbx+sbh > size_x Then sbx = size_x/2 : sbh = sbx / 2
        If sby+sbv > size_y Then sby = size_y/2 : sbv = sby / 2
        cx = sbx : cy = sby   ' move cursor back
      Case 85   ' U fill selection with colour  (undo)
        tmp = GetHexDialog("Fill selection colour: (0-F)")
        If tmp > -1 Then FillSelection(tmp) ' not cancel
      Case 148  ' F4  clear animation limits
        abx = 0 : aby = 0 : aex = size_x : aey = size_y
      Case 149  ' F5  mark start of animation
        abx = sbx : aby = sby
      Case 150  ' F6  move selection 1 width left
        StepSelectBackward
      Case 151  ' F7  move selection 1 width right
        StepSelectForward
      Case 152  ' F8  mark end of animation
        aex = sbx : aey = sby
      Case 153  ' F9  save current screen
        'Save image "screenshot"
      Case  3   ' ^C  Copy
        CopySelection
      Case 14   ' ^N  New   (undo)
        ClearWorksheet
      Case 15   ' ^O  Open  (undo)
        OpenWorksheet
      Case 17   ' ^Q  Quit - hope you saved your work!
        MODE 1
        End
      Case 19   ' ^S  Save
        SaveWorksheet
      Case 22   ' ^V  Paste (undo)
        PasteSelection
      Case 24   ' ^X  Cut   (undo)
        CopySelection
        FillSelection(0)
      Case 26   ' ^Z  Undo (after New,Open,Paste,Cut)
        If undoready Then
          Memory unpack Peek(varaddr undo1()),Peek(varaddr img()),4096,4
          undoready = 0
          ShowZoomView
        EndIf
    End Select
    ShowCurCol(img(cx,cy))
    ShowSSB
    ShowCur
    ShowValues
    ShowRealSize ' for speed, this should only show changes
  EndIf
Loop

' Copy the selection region to the CCP buffer
' Record the size of the region for later
Sub CopySelection
  Local x,y,t,c

  ccpx = sbh : ccpy = sbv   ' save size so later paste is correct
  For y = 0 To sbv - 1
    c = -1 : t = 0   ' keep track of nibbles in (64 bit) integers in array
    For x = 0 To sbh - 1
      Inc t,-1
      If t < 0 Then
        t = 15    ' 16 pixels (4 bits each) fit into one integer
        Inc c
        'Print c,y " ";
        'Print Hex$(ccp(c,y),16) " ";
        ccp(c,y) = 0
      EndIf
      ccp(c,y) = ccp(c,y) Or (img(x+sbx,y+sby) << t*4)
    Next x
    'If t < 15 Then Print Hex$(ccp(c,y),16);
    '?
  Next y
End Sub

' If there's something in the CCP buffer, paste it into the selection region.
' If the selection region is larger, paste into the top left
' If it's smaller, paste only the size of the region
Sub PasteSelection
  Local x,y,t,c

  If ccpx = 0 Or ccpy = 0 Then Exit Sub   ' nothing to paste
  SetUndo      ' make copy for undo
  For y = 0 To Min(ccpy,sbv) - 1
    c = 0 : t = 15
    For x = 0 To Min(ccpx,sbh) - 1
      img(x+sbx,y+sby) = (ccp(c,y) >> t*4) And 15  ' isolate 4 bits
      Inc t,-1
      If t < 0 Then
        t = 15
        Inc c
      EndIf
    Next x
  Next y
  ShowZoomView
End Sub

' Move the selection box backward by its width, wrap
'         abx,aby  aex,aey
Sub StepSelectBackward
  HideSSB
  If (sbx = abx) And (sby = aby) Then  ' at first frame
    sbx = aex : sby = aey   ' reset to ending
    ' clamp within safe limits
    If sbx+sbh >= size_x Then sbx = (size_x \ sbh - 1) * sbh
    If sby+sbv >= size_y Then sby = (size_y \ sbv - 1) * sbv
    Exit Sub
  EndIf
  Inc sbx,-sbh
  If sbx < 0 Then   ' horizontal wrap
    sbx = (size_x \ sbh - 1) * sbh
    Inc sby,-sbv
    If sby < 0 Then sby = (size_y \ sbv - 1) * sbv  ' vertical wrap
  EndIf
End Sub

' Move the selection box forward by its width, wrap
'         abx,aby  aex,aey
Sub StepSelectForward
  HideSSB
  If (sbx = aex) And (sby = aey) Then  ' at last frame
    sbx = abx : sby = aby   ' reset to begining
    Exit Sub
  EndIf
  Inc sbx,sbh
  If sbx+sbh > size_x Then   ' horizontal wrap
    sbx = 0
    Inc sby,sbv
    If sby+sbv > size_y Then sby = 0  ' vertical wrap
  EndIf
End Sub

Sub ShowRealSize
  Local x,y

  FRAMEBUFFER write n
  Box 32+rsx-sbh/2-2,rsy-2,sbh+4,sbv+4,1,0
  Box 32+rsx-sbh/2-1,rsy-1,sbh+2,sbv+2,1,RGB(rust)
  For y = 0 To sbv-1
    For x = 0 To sbh-1
      Pixel 32+rsx-sbh/2+x,rsy+y,ctab(img(sbx+x,sby+y))
    Next x
  Next y

End Sub

' Show position values along bottom
Sub ShowValues
  FRAMEBUFFER write n

  Text 5,MM.VRes-4,"Cursor X,Y:",LB,9,,RGB(yellow)
  Text 93,MM.VRes-4," "+Str$(cx)+",",RB,9,,RGB(yellow)
  Text 93,MM.VRes-4,Str$(cy)+" ",LB,9,,RGB(yellow)

  Text 125,MM.VRes-4,"Selection X,Y:",LB,9,,RGB(rust)
  Text 233,MM.VRes-4," "+Str$(sbx)+",",RB,9,,RGB(rust)
  Text 233,MM.VRes-4,Str$(sby)+" ",LB,9,,RGB(rust)

  Text 255,MM.VRes-4,"H,V:",LB,9,,RGB(rust)
  Text 300,MM.VRes-4," "+Str$(sbh)+",",RB,9,,RGB(rust)
  Text 300,MM.VRes-4,Str$(sbv)+" ",LB,9,,RGB(rust)

  Text 47,210,"Help  ?",RT,9
  RBox 37,207,13,13,3
End Sub

' Where's the sprite selction box?
Sub ShowSSB
  FRAMEBUFFER write n
  Box zoomx+sbx*sc,zoomy+sby*sc,sbh*sc+1,sbv*sc+1,1,RGB(rust)
End Sub

Sub HideSSB
  FRAMEBUFFER write n
  Box zoomx+sbx*sc,zoomy+sby*sc,sbh*sc+1,sbv*sc+1,1,RGB(myrtle)
End Sub

' Indicate which colour is under the cursor
Sub ShowCurCol(n)
  FRAMEBUFFER write n
  Box 19,25,8,95,1,0,0  ' erase old
  Box 54,25,8,95,1,0,0  ' erase old

  If n < 8 Then
    Circle 22,28+n*10,3,2,1,RGB(white)
  Else
    Circle 57,28+(n-8)*10,3,2,1,RGB(white)
  EndIf
End Sub

' Show the colour pallet on the left edge of the screen
Sub ShowPallet
  Local i

  FRAMEBUFFER write n
  For i = 0 To 7
    Text 1,25+i*10,Mid$(cn$,i+1,1),,9
    Box 8,23+i*10,10,10,1,RGB(white),ctab(i)
    Text 36,25+i*10,Mid$(cn$,i+8+1,1),,9
    Box 43,23+i*10,10,10,1,RGB(white),ctab(i+8)
  Next i
End Sub

Sub ShowCur
  FRAMEBUFFER write l
  Box zoomx+cx*sc,zoomy+cy*sc,sc+1,sc+1,1,RGB(yellow)
End Sub

Sub HideCur
  FRAMEBUFFER write l
  Box zoomx+cx*sc,zoomy+cy*sc,sc+1,sc+1,1,RGB(lilac)
End Sub

Sub DrawCell(x,y)
  Box zoomx+x*sc,zoomy+y*sc,sc+1,sc+1,1,RGB(myrtle),ctab(img(x,y))
End Sub

Sub ShowZoomView
  Local x,y
  FRAMEBUFFER write n
  For y = 0 To size_y-1
    For x = 0 To size_x-1
      Box zoomx+x*sc,zoomy+y*sc,sc+1,sc+1,1,RGB(myrtle),ctab(img(x,y))
    Next
  Next
End Sub

' Save data for possible undo
' Copy img() to undo(), packing bytes as we go
Sub SetUndo
  Memory pack Peek(varaddr img()),Peek(varaddr undo1()),4096,4
  undoready = 1
End Sub

' Clear entire worksheet, colour can be chosen
' allows for possible undo
Sub ClearWorksheet
  Local bg

  bg = GetHexDialog("New - select colour: (0-F)")
  If bg < 0 Then Exit Sub   ' cancel
  If bg > 15 Then bg = 0    ' black if out of range
  SetUndo      ' make copy for undo
  Math set bg,img()
  ShowZoomView
End Sub

' Sets all pixels in selection window to a colour
' allows for possible undo
Sub FillSelection(bg)
  Local x,y

  bg = Max(0,Min(bg,15))    ' black if out of range
  SetUndo      ' make copy for undo
  For y = sby To sby + sbv - 1
    For x = sbx To sbx + sbh - 1
      img(x,y) = bg
    Next x
  Next y
  ShowZoomView
End Sub

' flip the selection region horizontally (vertical axis)
' redraw affected cells
Sub FlipHorizontal
  Local x,y,t,x1,x2
  FRAMEBUFFER write n
  For y = sby To sby + sbv - 1
    For x = 0 To (sbh - 1) \ 2
      x1 = sbx+x
      x2 = sbx+sbh-x-1
      t = img(x1,y)
      img(x1,y) = img(x2,y)
      img(x2,y) = t
      Box zoomx+x1*sc,zoomy+y*sc,sc+1,sc+1,1,RGB(myrtle),ctab(img(x1,y))
      Box zoomx+x2*sc,zoomy+y*sc,sc+1,sc+1,1,RGB(myrtle),ctab(img(x2,y))
    Next x
  Next y
End Sub

' flip the selection region vertically (horizontal axis)
' redraw affected cells
Sub FlipVertical
  Local x,y,t,y1,y2
  FRAMEBUFFER write n
  For x = sbx To sbx + sbh - 1
    'For x = 0 To (sbh - 1) \ 2
    For y = 0 To (sbv - 1) \ 2
      y1 = sby+y
      y2 = sby+sbv-y-1
      t = img(x,y1)
      img(x,y1) = img(x,y2)
      img(x,y2) = t
      Box zoomx+x*sc,zoomy+y1*sc,sc+1,sc+1,1,RGB(myrtle),ctab(img(x,y1))
      Box zoomx+x*sc,zoomy+y2*sc,sc+1,sc+1,1,RGB(myrtle),ctab(img(x,y2))
    Next y
  Next x
End Sub

' flip the upper-left square of the selection region diagonally around
' a line from top left to bottom right
Sub FlipDiagonal
  Local x,y,t
  FRAMEBUFFER write n
  For y = 1 To Min(sbh,sbv) - 1
    For x = 0 To y-1
      x1 = sbx+x : y1 = sby+y
      x2 = sbx+y : y2 = sby+x
      t = img(x1,y1)
      img(x1,y1) = img(x2,y2)
      img(x2,y2) = t
      Box zoomx+x1*sc,zoomy+y1*sc,sc+1,sc+1,1,RGB(myrtle),ctab(img(x1,y1))
      Box zoomx+x2*sc,zoomy+y2*sc,sc+1,sc+1,1,RGB(myrtle),ctab(t)
    Next x
  Next y
End Sub

Sub MakeCSub
  Local c,i,j,x,y
  Local r$

  Print Hex$(sbv,4) Hex$(sbh,4)   ' size header, uncompressed
  c = 1 : i = 0 : j = 0 : r$ = ""
  For y = sby To sby + sbv - 1
    For x = sbx To sbx + sbh - 1
      r$ = Hex$(cmap(img(x,y)),1) + r$
      Inc i
      If i > 7 Then
        Print r$ " ";
        Inc c
        i = 0 : r$ = ""
        Inc j
        If j > 3 Then
          Print   ' 4 words per line
          j = 0
        EndIf
      EndIf
    Next x
  Next y
  If r$ <> "" Then
    Print Right$(String$(8,"0")+r$,8)
    Inc c
  EndIf
  Print "' length:" c*4

End Sub

' Show rudimentary help
Sub ShowHelp
  Local i
  HideCur
  FRAMEBUFFER write l
  RBox hbx-1,hby-1,hbw+10,hbh+10,,d_shadow, d_shadow
  RBox hbx-5,hby-5,hbw+10,hbh+10,,d_frame, d_frame
  RBox hbx,hby,hbw,hbh,5,d_back, d_back  ' text area
  Text hbx+hbw-10,hby+hbh-10,"Press a key...",RT,8
  DrawTitle(hby + 5)

  i = hby + 25
  ShowText(i,"Arrow keys move cursor",RGB(yellow))
  ShowText(i,"0-9 A-F set colour at cursor",RGB(yellow))
  Inc i,5
  ShowText(i,"I-J-K-L move selection region",RGB(rust))
  ShowText(i,"  , .   change selection width",RGB(rust))
  ShowText(i,"  P ;   change selection height",RGB(rust))
  ShowText(i,"M  flip selection horizontally",RGB(rust))
  ShowText(i,"H  flip selection vertically",RGB(rust))
  ShowText(i,"N  flip selection diagonally (square)",RGB(rust))
  ShowText(i,"U  fill selection with colour",RGB(rust))
  Inc i,5
  ShowText(i,"R  resize worksheet (no erase)")
  Inc i,5
  ShowText(i,"Ctrl-X/C/V  cut/copy/paste selection")
  ShowText(i,"Ctrl-O/S/N  open/save/clear worksheet")
  ShowText(i,"Ctrl-Z  undo (open/cut/paste/clear/fill)")
  Inc i,5
  ShowText(i,"G  generate CSub output to console")
  Inc i,5
  ShowText(i,"Ctrl-Q  quit - no warning!",RGB(red))

  Do : Loop Until Inkey$ <> ""

  Box hbx+20,hby+24,hbw-30,hbh-40,,0,0  ' erase page 1

  i = hby + 25
  ShowText(i,"     Simple Animation Features")
  Inc i,5
  ShowText(i,"Simple keystrokes allow quick steps")
  ShowText(i,"between sprites to check animation. The")
  ShowText(i,"selection region can be moved forward")
  ShowText(i,"and backward in the worksheet according")
  ShowText(i,"to its size. The region will wrap at")
  ShowText(i,"the edges of the screen.")
  Inc i,5
  ShowText(i,"F4  clear endpoints")
  ShowText(i,"F5  mark starting endpoint")
  ShowText(i,"F6  step backward one region")
  ShowText(i,"F7  step forward one region")
  ShowText(i,"F8  mark finishing endpoint")
  Inc i,5
  ShowText(i,"Move the region to the first frame and")
  ShowText(i,"press F5 to mark it. Use F7 to advance")
  ShowText(i,"the region to the last frame. Mark with")
  ShowText(i,"F8. Now use F6 & F7 to cycle.")

  Do : Loop Until Inkey$ <> ""

  CLS RGB(lilac)
  ShowCur
End Sub

Sub ShowText(y,t$,c)
  If c = 0 Then c = RGB(white)
  Text hbx+20,y,t$,LT,9,,c
  Inc y,10
End Sub

Sub DrawTitle(y)
  Text hbx+hbw/2,y+3,"Simple Sprite Draw",CT,1,,RGB(red)
  Text hbx+hbw/2-1,y+2,"Simple Sprite Draw",CT,1,,RGB(rust),-1
  Text hbx+hbw/2-2,y+1,"Simple Sprite Draw",CT,1,,RGB(brown),-1
  Text hbx+hbw/2-3,y,"Simple Sprite Draw",CT,1,,RGB(yellow),-1
  Text hbx+hbw-10,y+6,"by Vegipete",RT,8,,RGB(green)
End Sub

' let user enter a hex value, flash cursor at x,y
Function GetHexDialog(t$)
  Local w,x,y,k,tm

  HideCur
  w = 6 * Len(t$) + 20
  x = MM.HRes/2 : y = MM.VRes/2
  RBox x-w/2- 6,y- 6,w+20,20,,d_shadow, d_shadow
  RBox x-w/2-10,y-10,w+20,20,,d_frame, d_frame
  RBox x-w/2- 5,y- 6,w+10,12,5,d_back, d_back  ' text area
  Text x-10,y,t$,CM,9

  Do
    tm = (Timer \ 750) And 1
    k = Asc(UCase$(Inkey$))
    Select Case k
      Case 27   ' Esc
        GetHexDialog = -1
        Exit
      Case 48 To 57  ' digits 0 to 9
        GetHexDialog = k-48
        Exit
      Case 65 To 70  ' letters A to F
        GetHexDialog = k-55
        Exit
    End Select
    Box x+w/2-12,y-4,5,8,0,0,tm*RGB(white)
  Loop
  CLS RGB(lilac)
  ShowCur
End Function

' let user chose worksheet size
'   s = current selection
Function GetSizeDialog(s)
  Local k,up

  t$ = "Select worksheet size..."
  HideCur

    ' dialog box dimensions
  d_height = 49 + 1 * 8
  d_width = 235
  d_x = (MM.HRes - d_width)/2
  d_y = (MM.VRes - d_height)/2

  ' prepare layer
  FRAMEBUFFER write L
  CLS RGB(lilac)

  ' draw dialog box
  RBox d_x+4,d_y+ 4,d_width- 8,d_height- 8,10,d_shadow, d_shadow
  RBox d_x  ,d_y   ,d_width- 8,d_height- 8,10, d_frame, d_frame
  RBox d_x+5,d_y+22,d_width-18,d_height-34, 5,  d_back, d_back  ' text area
  Text d_x+10,d_y+7,t$, "LT",9,1,d_fore,-1
  Text d_x+d_width-15,d_y+5,"ABC", "RT",10,1,d_fore,-1

  up = 1  ' up date dialog?

  Do
    If up Then
      ShowSizes(s,d_x+5+(d_width-18)/2,d_y+22+(d_height-34)/2)
      up = 0
    EndIf
    k = Asc(UCase$(Inkey$))
    Select Case k
      Case 13   ' Enter
        GetSizeDialog = s
        Exit
      Case 27   ' Escape
        GetSizeDialog = -1
        Exit
      Case 130  ' left
        up = 1
        Inc s,-2
        If s < 0 Then Inc s,4
      Case 131  ' right
        up = 1
        Inc s,2
        If s > 3 Then Inc s,-4
      Case 128  ' up
        up = 1
        Inc s,-1
        If s < 0 Then s = 3
      Case 129  ' down
        up = 1
        Inc s
        If s > 3 Then s = 0
    End Select
  Loop
  CLS RGB(lilac)
End Function

Sub ShowSizes(n,x,y)
  Local i
  For i = 0 To 3
    If n = i Then
      Text x+wssx(i),y,wss$(i),wssj$(i),9,1,d_fore,d_frame
    Else
      Text x+wssx(i),y,wss$(i),wssj$(i),9,1,RGB(white),d_back
    EndIf
  Next i
End Sub

'==========================================================
' Show a dialog box to request a file name
' Top line is prompt text, bottom line for user input
Function SaveFileDialog(t$) As string
  Local w,x,y,tm,k$,n$

  HideCur
  w = 6 * Len(t$) + 20
  x = MM.HRes/2 : y = MM.VRes/2
  RBox x-w/2- 6,y- 6,w+20,30,,d_shadow, d_shadow
  RBox x-w/2-10,y-10,w+20,30,,d_frame, d_frame
  RBox x-w/2- 5,y- 6,w+10,22,5,d_back, d_back  ' text area
  Text x-10,y,t$,CM,9

  n$ = ""
  Do
    tm = (Timer \ 750) And 1
    k$ = (Inkey$)
    If k$ <> "" Then
      If Asc(k$) = 27 Then  ' cancel
        SaveFileDialog = ""
        Exit
      EndIf
      If Len(n$) = 0 Then
        If Instr(ltr$,k$) Then  ' 1st character must be a letter
          n$ = k$
          Text x-w/2,y+10,n$,LM,9
        EndIf
      Else  ' already got something
        If (Asc(k$) = 8) Or (Asc(k$) = 130) Then
          Box x-w/2+6*Len(n$),y+6,5,8,0,0,0   ' erase possible cursor
          n$ = Left$(n$,Len(n$)-1)  ' chop off a character
          Text x-w/2,y+10,n$+" ",LM,9
        ElseIf (Instr(ltr$+dgt$+" _-()+",k$) <> 0) And (Len(n$) < 32) Then
          n$ = n$ + k$
          Text x-w/2,y+10,n$,LM,9
        ElseIf Asc(k$) = 13 Then
          SaveFileDialog = n$
          Exit
        EndIf
      EndIf
    EndIf
    Box x-w/2+6*Len(n$),y+6,5,8,0,0,tm*RGB(white)
  Loop
  CLS RGB(lilac)
  ShowCur
End Function

'==========================================================
' File Handling Routines
'==========================================================
' User can select something to open, ESC to cancel
' Could be a worksheet, an original CMM sprite file
' or CSub data.
' Allows for possible undo
Sub OpenWorksheet
  Local f$,t$,x,y,h,w

  f$ = GetFileName("",10)
  If f$ <> "" Then
    SetUndo      ' make copy for undo

    Select Case UCase$(Right$(f$,4))
      Case ".TXT", ".SWS"
        Open f$ For input As #4
        w = Val(ReadFileNotComment())  ' read width
        h = Val(ReadFileNotComment())  ' read height
        ReadFilePixelArray(w,h)
        Close #4
      Case ".SPR"   ' original CMM sprite file
        Open f$ For input As #4
        t$ = ReadFileNotComment()  ' read size, number of sprites
        ' sprite must be 16 pixels square, only reads first one
        If Val(t$) = 16 Then ReadFilePixelArray(16,16)
        Close #4
    End Select
    ShowZoomView
  EndIf
End Sub

' Read an array of pixels
' Size is (xs,ys)
' Leading spaces are converted to 0 (black)
' Trailing spaces are ignored - maybe don't want this?
Sub ReadFilePixelArray(xs,ys)
  Local x,y
  ' allow for smaller saved files
  For y = 0 To ys-1
    t$ = ReadFileNotComment()   ' read line of pixels
    If t$ = "-1" Then Exit Sub  ' no file left to get pixels from!
    ConvertLeadSpace(t$,"0")
    For x = 0 To Min(Len(t$)-1,xs-1)
      img(x,y) = Val("&H"+Mid$(t$,x+1,1))
    Next x
  Next y
End Sub

' Change leading spaces in the passed string to the passed character
Sub ConvertLeadSpace(s$,c$)
  Local i

  c$ = Left$(c$,1)
  If c$ = " " Then Exit Sub   ' bit pointless to convert spaces to spaces
  i = 1
  Do While Mid$(s$,i,1) = " "  ' is this char a space?
    MID$(s$,i,1) = c$  ' yes, change and keep going
    Inc i
  Loop
End Sub

' Read and return next non-comment line of file
Function ReadFileNotComment() As string
  Local s$,i

  Do
    If Eof(#4) Then
      ReadFileNotComment = "-1"   ' indicate nothing left
      Exit Function
    Else
      Line Input #4,s$
      i = 0
      Do
        Inc i
      Loop Until Mid$(s$,i,1) <> " "  ' find first non-space char
      If Mid$(s$,i,1) = "'" Then Continue do  ' ignore comment line
      ReadFileNotComment = s$   ' return line of text
      Exit Function
    EndIf
  Loop
End Function

' User types in file name to save
' In future, might want to remember previously used name.
' Might also want to see existing files.
' Note: maximum possible worksheet is saved, not just what's visible
' Hint: save files could be merged by manually shrinking them. They
' will load into the upper left corner.
Sub SaveWorksheet
  Local x,y,f$

  f$ = SaveFileDialog(" Save Worksheet - File Name...")
  If f$ <> "" Then
    f$ = f$ + ".sws"
    Open f$ For output As #4
    Print #4,Str$(MAX_X)   ' save width
    Print #4,Str$(MAX_Y)   ' save height
    For y = 0 To MAX_Y-1
      For x = 0 To MAX_X-1
        Print #4,Hex$(img(x,y),1);
      Next x
      Print #4
    Next y
    Close #4
  EndIf
End Sub

'*****************************************************************
' Function GetFileName(spec$,vsize)
'
'   version 1.0   Original release vegipete, Oct 2020
'   version 1.1   all navigation by arrow keys only,
'                   can return directory names too
'   version 1.2   remembers selection when moving back up directories,
'                   can specify file criteria, forcing selection type
'
' This funtion displays a centered dialog box on the screen, allows
' the user to choose a file and returns the full path of the chosen
' file. The underlying screen is restored when the dialog closes.
' UP and DOWN arrows to select, ENTER to choose selection
' ESC to cancel, LEFT arrow to go up directory
'
' Input:  spec$: allow file filtering
'         vsize: number of directory items to list vertically, default 8
'
' Output: string containing full path of file chosen, or "" if nothing
'
' The following global variables should be declared before use:
' CONST DIRCOUNT = 50   ' max number of sub-directories
' CONST FILCOUNT = 255  ' max number of files
' CONST FNAMELEN = 32   ' filename length
' dim dir_dirs$(DIRCOUNT) length FNAMELEN  ' store list of directories
' dim dir_fils$(FILCOUNT) length FNAMELEN  ' store list of files
' dim dir_hist$(DIRCOUNT) length 8 ' store directory number visited along path
' dim dir_notroot  ' used internally to indicate root directory or not
'
' Routines Used:  (included below)
'   sub ReadDir   ' reads current directory into the above arrays
'   sub ListDir(first, nlines, hilite)  ' shows a portion of current directory
'
Function GetFileName(spec$,vsize) As string
  ' dialog box dimensions
  d_lines = vsize
  If d_lines < 8 Then d_lines = 8
  d_height = 50 + (d_lines - 1) * 8 ' Height for font 7
  d_width = 235
  d_x = (MM.HRes - d_width)/2
  d_y = (MM.VRes - d_height)/2

  d_startdir$ = Cwd$      ' save starting directory
  For d_k = 1 To DIRCOUNT ' set all elements to 1 - 1st item selected
    dir_hist$(d_k) = "1,1"
  Next d_k
  dir_hist$(0) = "1"      ' initially at top directory level
  If d_startdir$ <> "A:/" Then  ' determine starting directory depth
    d_startdir$ = d_startdir$ + "/"
    For d_k = 1 To Len(d_startdir$)
      If Mid$(d_startdir$,d_k,1) = "/" Then
        dir_hist$(0) = Str$(Val(dir_hist$(0)) + 1) ' another level deeper
      EndIf
    Next d_k
  EndIf

  ' prepare layer
  HideCur
  FRAMEBUFFER write L
  CLS RGB(lilac)

  ' draw dialog box
  RBox d_x+4,d_y+ 4,d_width- 8,d_height- 8,10,d_shadow, d_shadow
  RBox d_x  ,d_y   ,d_width- 8,d_height- 8,10, d_frame, d_frame
  RBox d_x+5,d_y+22,d_width-18,d_height-34, 5,  d_back, d_back  ' text area
  If UCase$(spec$) = "<DIR>" Then
    Text d_x+10,d_y+7,"Select Directory...", "LT",9,1,d_fore,-1
  Else
    Text d_x+10,d_y+7,"Select File...", "LT",9,1,d_fore,-1
  EndIf
  Text d_x+d_width-15,d_y+5,"ABC", "RT",10,1,d_fore,-1

  '--------------------
  ReadDir(spec$,d_top_item,d_sel_item,d_top_last)
  ListDir(d_top_item, d_lines, d_sel_item)  ' populate the dialog box

  Do
    d_k = Asc(Inkey$)
    d_changed = 0
    Select Case d_k
      Case  27  ' ESC
        GetFileName$ = ""  ' Cancel so return blank
        Exit
      Case 128  ' UP arrow
        If d_sel_item = 1 Then  ' is the top item selected?
          If d_top_item > 1 Then  ' at top of list?
            d_top_item = d_top_item - 1  ' no so shift list up one
            d_changed = 1
          EndIf
        Else
          d_sel_item = d_sel_item - 1  ' shift selection up one
          d_changed = 1
        EndIf
      Case 129  ' DOWN arrow
        If d_sel_item = d_lines Then  ' is the bottom item selected?
          If d_top_item < d_top_last Then  ' at bottom of list?
            d_top_item = d_top_item + 1  ' no so shift list down one
            d_changed = 1
          EndIf
        ElseIf d_sel_item < Val(dir_dirs$(0)) + Val(dir_fils$(0)) Then
          ' don't shift down past last item
          d_sel_item = d_sel_item + 1  ' shift selection down one
          d_changed = 1
        EndIf
      Case 130  ' LEFT Arrow - directory up if not root
        If Cwd$ <> "A:/" Then ' in a sub-directory?
          Chdir ".."     'directory up chosen
          ReadDir(spec$,d_top_item,d_sel_item,d_top_last)
          dir_hist$(0) = Str$(Val(dir_hist$(0)) - 1)
          d_top_item = Val(Field$(dir_hist$(Val(dir_hist$(0))),1,","))
          d_sel_item = Val(Field$(dir_hist$(Val(dir_hist$(0))),2,","))
          d_changed = 1
        EndIf
      Case 131  ' RIGHT Arrow - directory down if directory selected
        d_chosen = d_top_item + d_sel_item - 1
        If d_chosen <= Val(dir_dirs$(0)) Then ' item number in directory range

          dir_hist$(Val(dir_hist$(0))) = Str$(d_top_item)+","+Str$(d_sel_item)
          'dir_hist$(dir_hist$(0)) = d_chosen    ' save selection number if we
          dir_hist$(0) = Str$(Val(dir_hist$(0)) + 1)

          If Right$(Cwd$,1) = "/" Then
            Chdir Cwd$ + dir_dirs$(d_chosen)  ' tunnel down a directory from r
          Else
            Chdir Cwd$ + "/" + dir_dirs$(d_chosen)  ' tunnel down a directory
          EndIf
          ReadDir(spec$,d_top_item,d_sel_item,d_top_last)
          d_changed = 1
        EndIf

      Case  13  ' ENTER - something has been selected
        d_chosen = d_top_item + d_sel_item - 1
        If d_chosen <= Val(dir_dirs$(0)) Then ' item number in directory range
          If UCase$(spec$) = "<DIR>" Then   ' was directory selection chosen?
            If Right$(Cwd$,1) = "/" Then
              GetFileName$ = Cwd$ + dir_dirs$(d_chosen) + "/"  ' directory at
            Else
              GetFileName$ = Cwd$ + "/" + dir_dirs$(d_chosen) + "/"   ' direct
            EndIf     ' Note: cwd$ returns all uppercase
            Exit
          EndIf
        Else    ' Yahoo! A filename has been chosen
          If UCase$(spec$) <> "<DIR>" Then   ' was other than directory select
            d_chosen = d_chosen - Val(dir_dirs$(0))
            If dir_fils$(d_chosen) <> "" Then  ' in case directory has no (spe
              If Right$(Cwd$,1) = "/" Then
                GetFileName$ = Cwd$ + dir_fils$(d_chosen)  ' filename at root
              Else
                GetFileName$ = Cwd$ + "/" + dir_fils$(d_chosen)  ' filename de
              EndIf     ' Note: cwd$ returns all uppercase
              Exit
            EndIf
          EndIf
        EndIf
    End Select
    If d_changed Then   ' something changed so redisplay directory list
      ListDir(d_top_item, d_lines, d_sel_item)
    EndIf
  Loop
  '--------------------

  ' restore original screen image
  CLS RGB(lilac)

  ' restore starting directory
  Chdir d_startdir$

End Function

'*****************************************************************
' Read directories and specified files in the current directory
Sub ReadDir(spec$,d_top_item,d_sel_item,d_top_last)
  Local item_cnt, i

  For i = 1 To DIRCOUNT
    dir_dirs$(i) = ""   ' clear the array
  Next i
  For i = 1 To FILCOUNT
    dir_fils$(i) = ""   ' clear the array
  Next i

  ' read directories first
  dir_dirs$(0) = ""  ' 0 items to begin
  item_cnt = 1
  dir_dirs$(item_cnt) = Left$(Dir$("*", DIR),FNAMELEN) ' WARNING - possible tr
  Do While dir_dirs$(item_cnt) <> "" And item_cnt < DIRCOUNT - 1
    If dir_dirs$(item_cnt) <> "." Then item_cnt = item_cnt + 1 ' ignore "."
    dir_dirs$(item_cnt) = Dir$()
  Loop
  If dir_dirs$(item_cnt) = "" Then item_cnt = item_cnt - 1

  ' Sort directories
  Sort dir_dirs$()    ' note:  "" < "A"
  ' shift non-blank entries to front of array
  For i = 1 To item_cnt
    dir_dirs$(i) = dir_dirs$(DIRCOUNT-item_cnt+i)
  Next i
  dir_dirs$(0) = Str$(item_cnt)   ' store number of items

  ' now read files
  dir_fils$(0) = ""  ' 0 items to begin
  item_cnt = 1
  If UCase$(spec$) = "<DIR>" Then
    dir_fils$(item_cnt) = Left$(Dir$("*", FILE),FNAMELEN) ' WARNING - possible
  Else
    dir_fils$(item_cnt) = Left$(Dir$(spec$, FILE),FNAMELEN) ' WARNING - possib
  EndIf
  Do While dir_fils$(item_cnt) <> "" And item_cnt < FILCOUNT - 1
    If dir_fils$(item_cnt) <> "." Then item_cnt = item_cnt + 1 ' ignore "."
    dir_fils$(item_cnt) = Dir$()
  Loop
  If dir_fils$(item_cnt) = "" Then item_cnt = item_cnt - 1

  ' Sort files and shift non-blank entries to front of array
  Sort dir_fils$()
  For i = 1 To item_cnt
    dir_fils$(i) = dir_fils$(FILCOUNT-item_cnt+i)
  Next i
  dir_fils$(0) = Str$(item_cnt)   ' store number of items

  d_top_item = 1
  d_sel_item = 1
  d_top_last = Val(dir_dirs$(0)) + Val(dir_fils$(0)) - d_lines + 1

End Sub

'*****************************************************************
' Display (part of) directory
' Show 'nlines' number of items, starting with item 'first',
' hilite given item
' The magic number "33" should be altered to correspond to the magic
' number "300" used above for d_width.
Sub ListDir(first, nlines, hilite)
  Local i

  For i = 0 To nlines - 1
    item = first + i
    If item > Val(dir_dirs$(0)) Then
      d_txt$ = dir_fils$(item - Val(dir_dirs$(0)))
    Else
      d_txt$ = "<DIR> " + dir_dirs$(item)
    EndIf
    If Len(d_txt$) > 33 Then d_txt$ = Left$(d_txt$,32) + Chr$(148)
    d_txt$ = Left$(d_txt$ + Space$(33),33)

    If i = hilite - 1 Then
      Text d_x+14,d_y+24+i*MM.Info(FONTHEIGHT),d_txt$,"LT",9,1,d_fore,d_frame
    Else
      Text d_x+14,d_y+24+i*MM.Info(FONTHEIGHT),d_txt$,"LT",9,1,&hFFFFFF,d_back
    EndIf
  Next i

End Sub
'*****************************************************************

DefineFont #9
  60200806
  00000000 82200000 00022008 00004051 4F510000 0045F994 2B1CEA21 21C300C2
  80614208 AA108A42 80208046 00000000 40108410 81400081 00841004 7208A722
  820000A2 0080203E 01000000 00001082 0000003E 01000000 21000086 00004208
  8AAA2872 82210027 00872008 428C2072 21F8800F 00270A0C F824C510 0FFA0041
  00270A02 8A3C0831 20F80027 00822084 8A9C2872 28720027 0046089E 01180600
  06000086 10820118 40208410 0F000081 0000F880 10028140 21720084 00022008
  B2AE2A72 48218007 8028FAA2 8ABC28F2 2872002F 00278220 8AA228F2 08FA002F
  800F823C 823C08FA 087A0008 80279A20 8ABE288A 82708028 00872008 0A822008
  4A8A0027 8048A230 82200882 6A8B800F 80288AAA 9AAA2C8A 28728028 00278AA2
  82BC28F2 28720008 8046AAA2 A2BC28F2 28728048 00270A1C 200882F8 288A0082
  00278AA2 89A2288A 288A0042 8068ABAA 5208258A 258A8028 00822008 420821F8
  0CFB800F 800FC330 10080402 61F80020 806F1886 00228500 00000000 3E000000
  00008140 07000000 80277A02 8A220F82 0700002F 800782A0 8AA22708 07008027
  0007FA22 413C2431 07000004 9CE08922 8A220F82 06208028 00872008 10040310
  08821849 8048E2A4 20088260 0D000087 80A8AAAA 8A220F00 07008028 00278A22
  8B220F00 070020C8 82E089A2 82B00B00 07000008 002F70A0 41100F41 08000023
  80668AA2 89A20800 08000042 00A5AAA2 21940800 08008048 9CE089A2 21840F00
  8639800F 80836130 20088220 C3E00082 00CE3006 0000C06A A5020000 00A0522A
End DefineFont

DefineFont #10
  03410C14
  0E004000 00F00100 43182084 18F8F103 20840043 0E00F001 00400000 FF3F0000
  5E0200C4 22002502 CDA5A750 25925022 915E2209 3F0200A4 0000C0FF 00C4FF3F
  25005E02 73500200 5042C8A5 42012564 00A4E35E C0FF3F02 00000000
End DefineFont
